from argparse import Namespace

from .plugin import *


class Handle:
    @classmethod
    def ls(cls, args: Namespace) -> str:
        message = ""

        if args.store:
            if not args.is_superuser:
                return "获取插件商店需要超级用户权限！"
            message = "插件商店：\n"
            plugin = get_store_plugin_list()
        else:
            if args.conv["group"]:
                args.conv["user"] = []
            elif args.is_superuser:
                args.conv["user"] = []

            if args.user or args.group:
                if args.is_superuser:
                    args.conv = {"user": args.user, "group": args.group}
                else:
                    return "获取指定会话的插件列表需要超级用户权限！"

            for t in args.conv:
                for i in args.conv[t]:
                    message = f"{'用户' if t == 'user' else '群'} {i} 的插件列表：\n"

            plugin = plugin_manager.get_plugin(args.conv, 1)
            if not args.all:
                plugin = {
                    p: plugin[p]
                    for p in plugin
                    if plugin_manager.get_plugin(args.conv, 4)[p]
                }

        message = message + "\n".join(
            f"[{'o' if plugin[p] else 'x'}] {p}" for p in plugin
        )
        return message

    @classmethod
    def info(cls, args: Namespace) -> str:
        if not args.is_superuser:
            return "获取插件信息需要超级用户权限！"
        return get_plugin_info(args.plugin)

    @classmethod
    def chmod(cls, args: Namespace) -> str:
        if not args.is_superuser:
            return "设置插件权限需要超级用户权限！"
        plugin = plugin_manager.get_plugin()

        if args.all:
            args.plugin = list(plugin.keys())
        if args.reverse:
            args.plugin = list(filter(lambda p: p not in args.plugin, plugin))

        # TODO 这里之后应该有个将 r w x 翻译成 4 2 1 的处理
        result = plugin_manager.chmod_plugin(args.plugin, args.mode)

        return "\n".join(
            f"插件 {p} 的权限成功设置为 {args.mode}！" if result[p] else f"插件 {p} 不存在！"
            for p in result
        )

    @classmethod
    def block(cls, args: Namespace) -> str:
        if args.is_superuser:
            plugin = plugin_manager.get_plugin(perm=6)
        else:
            plugin = plugin_manager.get_plugin(conv=args.conv, perm=6)

        if args.conv["group"]:
            if not args.is_admin and not args.is_superuser:
                return "管理群插件需要群管理员权限！"
            args.conv["user"] = []

        if args.all:
            args.plugin = list(plugin.keys())
        if args.reverse:
            args.plugin = list(filter(lambda p: p not in args.plugin, plugin))

        result = {}
        for p in plugin:
            if p in args.plugin and not plugin[p]:
                args.plugin.remove(p)
                result[p] = False

        if args.user or args.group:
            if args.is_superuser:
                args.conv = {"user": args.user, "group": args.group}
            else:
                return "管理指定会话的插件需要超级用户权限！"

        result.update(plugin_manager.block_plugin(args.plugin, args.conv))

        message = ""
        for t in args.conv:
            if args.conv[t]:
                message += "用户 " if t == "user" else "群 "
                message += ",".join(str(i) for i in args.conv[t])
        message += " 中："

        for plugin, value in result.items():
            message += "\n"
            if value:
                message += f"插件 {plugin} 禁用成功！"
            else:
                message += f"插件 {plugin} 不存在或已关闭编辑权限！"
        return message

    @classmethod
    def unblock(cls, args: Namespace) -> str:
        if args.is_superuser:
            plugin = plugin_manager.get_plugin(perm=6)
        else:
            plugin = plugin_manager.get_plugin(conv=args.conv, perm=6)

        if args.conv["group"]:
            if not args.is_admin and not args.is_superuser:
                return "管理群插件需要群管理员权限！"
            args.conv["user"] = []

        if args.all:
            args.plugin = list(plugin.keys())
        if args.reverse:
            args.plugin = list(filter(lambda p: p not in args.plugin, plugin))

        result = {}
        for p in plugin:
            if p in args.plugin and not plugin[p]:
                args.plugin.remove(p)
                result[p] = False

        if args.user or args.group:
            if args.is_superuser:
                args.conv = {"user": args.user, "group": args.group}
            else:
                return "管理指定会话的插件需要超级用户权限！"

        result.update(plugin_manager.unblock_plugin(args.plugin, args.conv))

        message = ""
        for t in args.conv:
            if args.conv[t]:
                message += "用户" if t == "user" else "群"
                message += ",".join(str(i) for i in args.conv[t])
        message += "中："

        for plugin, value in result.items():
            message += "\n"
            if value:
                message += f"插件 {plugin} 启用成功！"
            else:
                message += f"插件 {plugin} 不存在或已关闭编辑权限！"
        return message

    # 以下功能尚未实现

    @classmethod
    def install(cls, args: Namespace) -> str:
        return ""

    @classmethod
    def uninstall(cls, args: Namespace) -> str:
        return ""



from argparse import Namespace

from nonebot.matcher import Matcher
from nonebot.params import ShellCommandArgs
from nonebot.message import run_preprocessor
from nonebot.exception import IgnoredException
from nonebot.plugin import on_shell_command, get_loaded_plugins
from nonebot.adapters.onebot.v11 import Bot, Event, MessageEvent, GroupMessageEvent

from .handle import Handle
from .parser import npm_parser
from .manager import plugin_manager

npm = on_shell_command("npm", parser=npm_parser, priority=1)

# 在 Matcher 运行前检测其是否启用
@run_preprocessor
async def _(matcher: Matcher, bot: Bot, event: Event):
    plugin = matcher.plugin_name

    conv = {
        "user": [event.user_id] if hasattr(event, "user_id") else [],  # type: ignore
        "group": [event.group_id] if hasattr(event, "group_id") else [],  # type: ignore
    }

    if (
        hasattr(event, "user_id")
        and not hasattr(event, "group_id")
        and str(event.user_id) in bot.config.superusers  # type: ignore
    ):
        conv["user"] = []
        conv["group"] = []

    plugin_manager.update_plugin(
        {
            str(p.name): p.name != "nonebot_plugin_manager" and bool(p.matcher)
            for p in get_loaded_plugins()
        }
    )

    if plugin and not plugin_manager.get_plugin(conv=conv, perm=1)[plugin]:
        raise IgnoredException(f"Nonebot Plugin Manager has blocked {plugin} !")


@npm.handle()
async def _(bot: Bot, event: MessageEvent, args: Namespace = ShellCommandArgs()):
    args.conv = {
        "user": [event.user_id],
        "group": [event.group_id] if isinstance(event, GroupMessageEvent) else [],
    }
    args.is_admin = (
        event.sender.role in ["admin", "owner"]
        if isinstance(event, GroupMessageEvent)
        else False
    )
    args.is_superuser = str(event.user_id) in bot.config.superusers

    if hasattr(args, "handle"):
        message = getattr(Handle, args.handle)(args)
        if message is not None:
            message = message.split("\n")
            if len(message) > 30:
                i = 1
                messages = []
                while len(message) > 30:
                    messages.append("\n".join(message[:30]) + f"\n【第{i}页】")
                    message = message[31:]
                    i = i + 1
                messages.append("\n".join(message[:30]) + f"\n【第{i}页-完】")
                if isinstance(event, GroupMessageEvent):
                    await bot.send_group_forward_msg(
                        group_id=event.group_id,
                        messages=[
                            {
                                "type": "node",
                                "data": {
                                    "name": "NBPM",
                                    "uin": bot.self_id,
                                    "content": msg,
                                },
                            }
                            for msg in messages
                        ],
                    )
                else:
                    await bot.send_private_forward_msg(
                        user_id=event.user_id,
                        messages=[
                            {
                                "type": "node",
                                "data": {
                                    "name": "NBPM",
                                    "uin": bot.self_id,
                                    "content": msg,
                                },
                            }
                            for msg in messages
                        ],
                    )
            else:
                await bot.send(event, "\n".join(message[:30]))



from nonebot.rule import ArgumentParser

npm_parser = ArgumentParser("npm")

npm_subparsers = npm_parser.add_subparsers(dest="handle")

list_parser = npm_subparsers.add_parser("ls", help="show plugin list")
list_group = list_parser.add_mutually_exclusive_group()
list_group.add_argument("-s", "--store", action="store_true")
list_group.add_argument(
    "-u", "--user", action="append", nargs="?", default=[], type=int
)
list_group.add_argument(
    "-g", "--group", action="append", nargs="?", default=[], type=int
)
list_parser.add_argument("-a", "--all", action="store_true")

info_parser = npm_subparsers.add_parser("info", help="show plugin info")
info_parser.add_argument("plugin", help="plugin you want to know about")

chmod_parser = npm_subparsers.add_parser("chmod", help="set plugin mode")
chmod_parser.add_argument("mode", type=str, help="mode you want to set")
chmod_parser.add_argument("plugin", nargs="*", help="plugin you want to set")
chmod_parser.add_argument("-a", "--all", action="store_true")
chmod_parser.add_argument("-r", "--reverse", action="store_true")

block_parser = npm_subparsers.add_parser("block", help="block plugin")
block_parser.add_argument("plugin", nargs="*", help="plugins you want to block")
block_parser.add_argument("-a", "--all", action="store_true")
block_parser.add_argument("-r", "--reverse", action="store_true")
block_parser.add_argument(
    "-u", "--user", action="store", nargs="+", default=[], type=int
)
block_parser.add_argument(
    "-g", "--group", action="store", nargs="+", default=[], type=int
)

unblock_parser = npm_subparsers.add_parser("unblock", help="unblock plugin")
unblock_parser.add_argument("plugin", nargs="*", help="plugins you want to unblock")
unblock_parser.add_argument("-a", "--all", action="store_true")
unblock_parser.add_argument("-r", "--reverse", action="store_true")
unblock_parser.add_argument(
    "-u", "--user", action="store", nargs="+", default=[], type=int
)
unblock_parser.add_argument(
    "-g", "--group", action="store", nargs="+", default=[], type=int
)

# 以下功能尚未实现

install_parser = npm_subparsers.add_parser("install", help="install plugin")
install_parser.add_argument("plugins", nargs="*", help="plugins you want to install")
install_parser.add_argument("-i", "--index", action="store", help="point to a mirror")

uninstall_parser = npm_subparsers.add_parser("uninstall", help="uninstall plugin")
uninstall_parser.add_argument(
    "plugins", nargs="*", help="plugins you want to uninstall"
)



from typing import Optional

try:
    from pip._internal.cli.main import main as pipmain
except ImportError:
    from pip import main as pipmain


def _call_pip_install(package: str, index: Optional[str] = None):
    if index:
        return pipmain(["install", "-i", index, package])
    else:
        return pipmain(["install", package])


def _call_pip_update(package: str, index: Optional[str] = None):
    if index:
        return pipmain(["install", "--upgrade", "-i", index, package])
    else:
        return pipmain(["install", "--upgrade", package])


def _call_pip_uninstall(package: str):
    return pipmain(["uninstall", package])



from typing import Dict

import httpx

from .manager import plugin_manager


# 获取商店插件列表
def __get_store_plugin_list() -> dict:
    return {
        plugin["module_name"]: plugin
        for plugin in httpx.get(
            "https://cdn.jsdelivr.net/gh/nonebot/nonebot2/website/static/plugins.json"
        ).json()
    }


def get_store_plugin_list() -> Dict[str, bool]:
    plugin_list = plugin_manager.get_plugin()
    return {plugin: plugin in plugin_list for plugin in __get_store_plugin_list()}


def get_plugin_info(plugin: str) -> str:
    store_plugin_list = __get_store_plugin_list()
    if plugin in store_plugin_list:
        plugin = store_plugin_list[plugin]
        return (
            f"ID: {plugin['module_name']}\n"
            f"Name: {plugin['name']}\n"
            f"Description: {plugin['desc']}\n"
            f"Latest Version: {httpx.get('https://pypi.org/pypi/'+plugin['project_link']+'/json').json()['info']['version']}\n"
            f"Author: {plugin['author']}\n"
            f"HomePage: {plugin['homepage']}"
        )
    else:
        return "查无此插件！"



from pathlib import Path
from typing import Dict, List, Union, Optional

import yaml

Conv = Dict[str, List[int]]


class PluginManager:
    __path: Path
    __plugin_list: Dict[str, Dict[str, Union[str, Dict[int, int]]]]

    def __init__(self, path: Path = Path() / "data" / "manager" / "plugin_list.yml"):
        self.__path = path
        self.__load()

    def get_plugin(
        self, conv: Conv = {"user": [], "group": []}, perm: Union[str, int] = 5
    ) -> Dict[str, bool]:
        """
        获取指定会话的插件列表

        参数:
            `conv: Conv` - 会话列表
            `perm: Union[str,int]` - 检测权限，满足该条件才返回 True

        返回:
            `Dict[str, bool]`
        """
        result = {}
        type = len(conv["user"]) + len(conv["group"])

        for p in self.__plugin_list:
            mode = int(self.__plugin_list[p]["mode"][type])
            for t in conv:
                can_break = False
                for i in conv[t]:
                    if i in self.__plugin_list[p][t]:
                        mode = int(self.__plugin_list[p][t][i])
                        if t == "user":
                            can_break = True
                if can_break:
                    break
            result[p] = mode & perm == perm
        return result

    # 设置插件模式
    def chmod_plugin(self, plugin: List[str], mode: str) -> Dict[str, bool]:
        result = {}
        for p in plugin:
            result[p] = False
            if p in self.__plugin_list:
                result[p] = True
                self.__plugin_list[p]["mode"] = mode
        self.__dump()
        return result

    # 禁用插件
    def block_plugin(
        self,
        plugin: List[str] = [],
        conv: Conv = {"user": [], "group": []},
    ) -> Dict[str, bool]:
        result = {}
        for p in plugin:
            result[p] = False
            if p in self.get_plugin():
                result[p] = True
                for t in conv:
                    for i in conv[t]:
                        if i in self.__plugin_list[p][t]:
                            mode = self.__plugin_list[p][t][i]
                        else:
                            mode = int(
                                self.__plugin_list[p]["mode"][1 if t == "user" else 2]
                            )
                        self.__plugin_list[p][t][i] = mode & 6

        self.__dump()
        return result

    # 启用插件
    def unblock_plugin(
        self,
        plugin: List[str] = [],
        conv: Conv = {"user": [], "group": []},
    ) -> Dict[str, bool]:
        result = {}
        for p in plugin:
            result[p] = False
            if p in self.get_plugin():
                result[p] = True
                for t in conv:
                    for i in conv[t]:
                        if i in self.__plugin_list[p][t]:
                            mode = self.__plugin_list[p][t][i]
                        else:
                            mode = int(
                                self.__plugin_list[p]["mode"][1 if t == "user" else 2]
                            )
                        self.__plugin_list[p][t][i] = mode | 1
        self.__dump()
        return result

    def update_plugin(self, plugin: Dict[str, bool]) -> "PluginManager":
        for p in plugin:
            if p not in self.__plugin_list:
                self.__plugin_list[p] = {
                    "mode": "755" if plugin[p] else "311",
                    "user": {},
                    "group": {},
                }
            elif plugin[p] ^ self.__plugin_list[p].get(
                "status",
                int(self.__plugin_list[p]["mode"][0]) & 4 == 4,
            ):
                self.__plugin_list[p]["mode"] = "".join(
                    str(int(m) | 4 if plugin[p] else int(m) & 3)
                    for m in self.__plugin_list[p]["mode"]
                )
            self.__plugin_list[p]["status"] = plugin[p]
        for p in self.__plugin_list:
            if p not in plugin:
                self.__plugin_list[p]["status"] = False
                self.__plugin_list[p]["mode"] = "".join(
                    str(int(m) & 3) for m in self.__plugin_list[p]["mode"]
                )
        return self

    # 加载插件列表
    def __load(self) -> "PluginManager":
        try:
            self.__plugin_list = yaml.safe_load(self.__path.open("r", encoding="utf-8"))
        except:
            self.__plugin_list = {}
        return self

    # 保存插件列表
    def __dump(self):
        self.__path.parent.mkdir(parents=True, exist_ok=True)
        yaml.dump(
            self.__plugin_list,
            self.__path.open("w", encoding="utf-8"),
            allow_unicode=True,
        )


plugin_manager = PluginManager()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_manager']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.1,<1.0.0',
 'nonebot-adapter-onebot>=2.0.0-rc.1,<3.0.0',
 'nonebot2>=2.0.0-rc.1,<3.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-manager',
    'version': '0.5.13',
    'description': 'Nonebot Plugin Manager base on import hook',
    'long_description': '<div align="center">\n  <img width="200" src="docs/logo.png" alt="logo"></br>\n\n# Nonebot Plugin Manager\n\n基于 [nonebot2](https://github.com/nonebot/nonebot2) 和 [go-cqhttp](https://github.com/Mrs4s/go-cqhttp) 的**非侵入式**插件管理器\n\n[![License](https://img.shields.io/github/license/Jigsaw111/nonebot_plugin_manager)](LICENSE)\n![Python Version](https://img.shields.io/badge/python-3.7.3+-blue.svg)\n![NoneBot Version](https://img.shields.io/badge/nonebot-2.0.0a11+-red.svg)\n![Pypi Version](https://img.shields.io/pypi/v/nonebot-plugin-manager.svg)\n\n</div>\n\n## 安装\n\n**插件仍在~~快速~~开发中，遇到问题还请务必提 issue。**\n\n### 从 PyPI 安装（推荐）\n\n- 使用 nb-cli  \n\n```bash\nnb plugin install nonebot-plugin-manager\n```\n\n- 使用 poetry\n\n```bash\npoetry add nonebot-plugin-manager\n```\n\n- 使用 pip\n\n```bash\npip install nonebot-plugin-manager\n```\n\n### 从 GitHub 安装（不推荐）\n\n```bash\ngit clone https://github.com/Jigsaw111/nonebot_plugin_manager.git\n```\n\n## 使用\n\n### 权限\n\n权限与 UNIX 的权限类似，分为三种用户：超级用户、用户、群。\n\n每种用户包含读、写、执行 3 个权限，分别对应数字 4、2、1，将 3 个权限对应的数字累加，最终得到的值即可作为每种用户所具有的权限。\n\n关于会话中使用什么模式，可参照下表：\n\n| 仅供参考 | 私聊读   | 群聊读 | 私聊写   | 群聊写    | 私聊执行 | 群聊执行  |\n| -------- | -------- | ------ | -------- | --------- | -------- | --------- |\n| 超级用户 | 超级用户 | 群     | 超级用户 | 超级用户  | 超级用户 | 群        |\n| 用户     | 用户     | 群     | 用户     | 无权限    | 用户     | 用户 & 群 |\n| 群管理员 | 不存在   | 群     | 不存在   | 用户 & 群 | 不存在   | 用户 & 群 |\n\n包含 Matcher 的插件默认权限为`755`，不含 Matcher 的插件默认权限为`311`。\n\n> 例：`npm chmod 757 nonebot_plugin_nodice`命令可将 nonebot_plugin_nodice 的权限设置为`757`\n> 即超级用户可写可读可执行，用户可读可执行，群可写可读可执行。\n\n只有超级用户可以修改插件的权限，可以使用绝对模式（八进制数字模式）~~，符号模式~~指定文件的权限。\n\n### 命令\n\n**使用前请先确保命令前缀为空，否则请在以下命令前加上命令前缀 (默认为`/`)。**\n\n- `npm ls`查看当前会话插件列表\n  - `-s, --store`互斥参数，查看插件商店列表（仅超级用户可用）\n  - `-u <user_id>, --user <user_id>`互斥参数，查看指定用户插件列表（仅超级用户可用）\n  - `-g <group_id>, --group <group_id>`互斥参数，查看指定群插件列表（仅超级用户可用）\n  - `-a, --all`可选参数，查看所有插件（包括不含 Matcher 的插件）\n\n- `npm info <插件名>`查询插件信息 （仅超级用户可用）\n\n- `npm chmod <mode> <plugin ...>`设置插件权限（仅超级用户可用）\n  - `mode`必选参数，需要设置的权限，参考上文\n  - `plugin ...`必选参数，需要设置的插件名\n  - `-a, --all`可选参数，全选插件\n  - `-r, --reverse`可选参数，反选插件\n\n- `npm block <plugin ...>`禁用当前会话插件（需要权限）\n  - `plugin ...`必选参数，需要禁用的插件名\n  - `-a, --all`可选参数，全选插件\n  - `-r, --reverse`可选参数，反选插件\n  - `-u <user_id ...>, --user <user_id ...>`可选参数，管理指定用户设置（仅超级用户可用）\n  - `-g <group_id ...>, --group <group_id ...>`可选参数，管理指定群设置（仅超级用户可用）\n\n- `npm unblock <plugin ...>`启用当前会话插件（需要权限）\n  - `plugin ...`必选参数，需要禁用的插件名\n  - `-a, --all`可选参数，全选插件\n  - `-r, --reverse`可选参数，反选插件\n  - `-u <user_id ...>, --user <user_id ...>`可选参数，管理指定用户设置（仅超级用户可用）\n  - `-g <group_id ...>, --group <group_id ...>`可选参数，管理指定群设置（仅超级用户可用）\n\n<!-- TODO\n\n- `npm install plugin...`安装插件（仅超级用户可用）\n  - `-i index, --index index`指定 PyPI 源\n\n- `npm uninstall plugin...`卸载插件（仅超级用户可用）\n  - `-a, --all`可选参数，全选插件\n\n-->\n\n## Q&A\n\n- **这是什么？**  \n  基于 import hook 的插件管理器，能够在不重启 NoneBot2 的情况下分群管理插件。\n- **有什么用？**  \n  在 NoneBot2 仍然缺乏插件管理机制的时期暂时充当插件管理器。\n- **自造 Rule 不是更好？**  \n  Rule 当然更好且更有效率，但是 Rule 是一种**侵入式**的插件管理方式，需要用户自行修改其他插件的源码，这对于管理从 PyPI 安装的插件来说相对复杂。而使用本插件，你不需要修改其他插件的任何内容，更符合插件之间**松耦合**的设计原则。\n\n<details>\n<summary>展开更多</summary>\n\n## 原理\n\n使用`run_preprocessor`装饰器，在 Matcher 运行之前检测其所属的 Plugin 判断是否打断。\n\n事实上 Nonebot 还是加载了插件，所以只能算是**屏蔽**而非**卸载**。\n\n<!-- TODO\n\n当然，你也可以使用`npm uninstall`命令来真正卸载插件，但我不建议你这样做，因为该命令将会重启 Nonebot 。\n\n-->\n\n## To Do\n\n- [x] 分群插件管理\n- [ ] 完善权限系统\n- [ ] 设置插件别名\n\n*咕咕咕*\n\n- [ ] 安装卸载插件\n\n## Bug\n\n- [ ] 无法停用 Matcher 以外的机器人行为（如 APScheduler ）  \n  **解决方法：** 暂无\n\n</details>\n',
    'author': 'Jigsaw',
    'author_email': 'j1g5aw@foxmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Jigsaw111/nonebot_plugin_manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
